"""Tests for querying data from Mixpanel."""

from pyramid_mixpanel.query import MixpanelQuery

import pytest
import responses

SETTINGS = {"mixpanel.api_secret": "bar"}


@responses.activate
def test_zero_results() -> None:
    """Return None if no profiles found."""
    responses.add(
        responses.POST, "https://mixpanel.com/api/2.0/jql", json=[], status=200
    )

    assert MixpanelQuery(SETTINGS).profile_by_email("foo") is None


@responses.activate
def test_too_many_results() -> None:
    """Raise exception if more than one profiles found."""
    responses.add(
        responses.POST,
        "https://mixpanel.com/api/2.0/jql",
        json=[
            {"distinct_id": "one", "email": "foo@bar.com"},
            {"distinct_id": "two", "email": "foo@bar.com"},
        ],
        status=200,
    )

    with pytest.raises(ValueError) as cm:
        MixpanelQuery(SETTINGS).profile_by_email("foo@bar.com")

    assert (
        str(cm.value)
        == "Found more than one Profile for email 'foo@bar.com': ['one', 'two']"
    )


@responses.activate
def test_profile_by_email() -> None:
    """Test happy path."""
    responses.add(
        responses.POST,
        "https://mixpanel.com/api/2.0/jql",
        json=[{"distinct_id": "foo", "email": "foo@bar.com"}],
        status=200,
    )

    assert MixpanelQuery(SETTINGS).profile_by_email("foo@bar.com") == {
        "distinct_id": "foo",
        "email": "foo@bar.com",
    }



"""Querying data from Mixpanel."""

import requests
import typing as t


class MixpanelQuery(object):
    """Query Mixpanel for events and profiles.

    You can use one of the pre-built queries, or provide your own.
    """

    ENDPOINT = "https://mixpanel.com/api/2.0/jql"

    def __init__(self, settings: t.Dict[str, str]) -> None:
        """Save API credentials."""
        self.api_secret = settings["mixpanel.api_secret"]

    def jql(self, jql: str) -> t.List[t.Dict]:
        """Query Mixpanel using JQL script.

        You can troubleshoot the script on
        https://mixpanel.com/report/<PROJECT_ID>/jql-console.
        """
        resp = requests.post(
            self.ENDPOINT, auth=(self.api_secret, ""), data={"script": jql}
        )
        resp.raise_for_status()
        return resp.json()

    #################################################
    #         ---- Pre-built queries ----           #
    # A couple of queries that we use all the time. #
    #################################################

    def profile_by_email(self, email: str):
        """Return a Mixpanel profile by given email."""
        profiles = self.jql(
            """
            function main() {
              return People(
              )
              .filter(function(profile) {
                return profile.properties.$email == '%(email)s';
              })
              .map(function(profile) {
                return {
                  distinct_id: profile.distinct_id,
                  email: profile.properties.$email,
                };
              });
            }
        """
            % {"email": email}
        )

        if len(profiles) == 0:
            return None
        elif len(profiles) == 1:
            return profiles[0]
        else:
            raise ValueError(
                f"Found more than one Profile for email '{email}': "
                f"{[profile['distinct_id'] for profile in profiles]}"
            )



"""Integration with Mixpanel.

There are two things you can do:
* Send user interaction events or set profile properties via
  track.MixpanelTrack().
* Query for events and profiles via query.MixpanelQuery().

MixpanelTrack is prepared as `request.mixpanel` for easier usage in view code.
"""
from dataclasses import dataclass
from pyramid.config import Configurator
from pyramid.events import NewRequest

import structlog

logger = structlog.get_logger(__name__)


@dataclass(frozen=True)
class Event:
    """A single event that we send to Mixpanel."""

    # The name of this event that will be shown in Mixpanel. Should be
    # something nice, like "Page Viewed".
    name: str


@dataclass(frozen=True)
class Events:
    """Let's be precise which events we will send.

    Otherwise different parts of code will send slightly differently named
    events and then you can wish good luck to the marketing/product team
    when they are trying to decipher what's the difference between
    "Page Load", "User Visited" and "Viewed Page".

    So we're only allowing properties listed below. You can provide your own
    list of properties via the `mixpanel.events` setting.
    """

    # Any page view should be tracked with this event. More specific events
    # can then be built with Mixpanel Web UI by filtering these events by
    # path and/or title. Further info:
    # https://help.mixpanel.com/hc/en-us/articles/115004562246-Combine-Events-To-Create-A-Custom-Event
    page_viewed: Event = Event("Page Viewed")

    # Any click should be tracked with this event. More specific events
    # can then be built with Mixpanel Web UI by filtering these events by
    # path and/or title. Further info:
    # https://help.mixpanel.com/hc/en-us/articles/115004562246-Combine-Events-To-Create-A-Custom-Event
    button_link_clicked: Event = Event("Button/Link Clicked")

    # Events for tracking user's account/subscription status.
    user_signed_up: Event = Event("User Signed Up")
    user_logged_in: Event = Event("User Logged In")
    user_charged: Event = Event("User Charged")
    user_disabled: Event = Event("User Disabled")


@dataclass(frozen=True)
class Property:
    """A single property that we attach to Mixpanel events or profiles."""

    # The name of this property that will be shown in Mixpanel. Should be
    # something nice, like "Path" or "Title". Some properties are ~special~
    # and they are prefixed with the dollar sign ($). Read more about them on
    # https://help.mixpanel.com/hc/en-us/articles/115004602703-Reserved-or-Special-Properties
    name: str


@dataclass(frozen=True)
class EventProperties:
    """Let's be precise which properties we will set on Events.

    Otherwise different parts of code will set slightly differently named
    properties and then you can wish good luck to the marketing/product team
    when they are trying to decipher what's the difference between "title" and
    "name".

    So we're only allowing properties listed below. You can provide your own
    list of properties via the `mixpanel.event_properties` setting.
    """

    # Used for building bespoke custom events from Page Viewed and
    # Button/Link Clicked events.
    title: Property = Property("Title")
    path: Property = Property("Path")

    # Referring URL, including your own domain.
    dollar_referrer: Property = Property("$referrer")


@dataclass(frozen=True)
class ProfileProperties:
    """Let's be precise which properties we will set on Profile records.

    Otherwise different parts of code will set slightly differently named
    properties and then you can wish good luck to the marketing/product team
    when they are trying to decipher what's the difference between "state" and
    "status".

    So we're only allowing properties listed below. You can provide your own
    list of properties via `mixpanel.profile_properties` setting.
    """

    # The time when the user created their account. This should be expressed
    # as a Mixpanel date string.
    dollar_created: Property = Property("$created")

    # The user's email address as a string, e.g. "joe.doe@example.com".
    # Mixpanel will use the "$email" property when sending email messages
    # to your users, and for displaying the user's gravatar image in reports.
    dollar_email: Property = Property("$email")

    # The users full name. If it is set, the full name of the user will be
    # displayed in Mixpanel reports, plus you can craft more personalized
    # automated email and in-app message.
    dollar_name: Property = Property("$name")

    # The user's phone number as a string, e.g. "4805551212". Mixpanel will
    # use the "$phone" property when sending SMS messages to your users.
    dollar_phone: Property = Property("$phone")

    # If this property is set to any value, a user will be unsubscribed
    # from Mixpanel automated email messages.
    dollar_unsubscribed: Property = Property("$unsubscribed")


@dataclass(frozen=True)
class ProfileMetaProperties:
    """Warning: here be dragons! Overrides of how Mixpanel works.

    There are used very rarely to send special values to Mixpanel to override
    sane default behavior.
    """

    # The IP address associated with a given profile. Mixpanel uses IP for
    # geo-locating the profile. If $ip isn't provided the underlying library
    # will set it to 0, and Mixpanel will ignore it. You probably have Mixpanel
    # installed in the frontend, and that sends over user's real IP so you can
    # leave this one empty.
    dollar_ip: Property = Property("$ip")

    # Seconds since midnight, January 1st 1970, UTC. Updates are applied
    # in $time order, so setting this value can lead to unexpected results
    # unless care is taken. If $time isn't provided the underlying library
    # will set it to `time.time()`.
    dollar_time: Property = Property("$time")

    # If the $ignore_time property is present and true in your update request,
    # Mixpanel will not automatically update the "Last Seen" property of the
    # profile. Otherwise, Mixpanel will add a "Last Seen" property associated
    # with the current time for all $set, $append, and $add operations.
    dollar_ignore_time: Property = Property("$ignore_time")

    # If the $ignore_alias property is present and true in your update request,
    # Mixpanel will apply the update directly to the profile with the
    # distinct_id included in the request, rather than allowing this
    # distinct_id to be recognized as an alias during ingestion.
    dollar_ignore_alias: Property = Property("$ignore_alias")


def includeme(config: Configurator) -> None:
    """Pyramid knob."""
    from pyramid_mixpanel.track import mixpanel_init
    from pyramid_mixpanel.track import mixpanel_flush
    from pyramid_mixpanel.track import MixpanelTrack
    from pyramid_mixpanel.consumer import MockedConsumer

    mixpanel = MixpanelTrack(settings=config.registry.settings)
    logger.info(
        "Mixpanel configured",
        consumer=mixpanel.api._consumer.__class__.__name__,
        events=mixpanel.events.__class__.__name__,
        event_properties=mixpanel.event_properties.__class__.__name__,
        profile_properties=mixpanel.profile_properties.__class__.__name__,
        profile_meta_properties=mixpanel.profile_meta_properties.__class__.__name__,
    )
    if mixpanel.api._consumer.__class__ == MockedConsumer:
        logger.warning("Mixpanel is in testing mode, no message will be sent!")

    config.add_request_method(mixpanel_init, "mixpanel", reify=True)
    config.add_subscriber(mixpanel_flush, NewRequest)



"""Consumers send events/profiles messages to Mixpanel's HTTP API."""

from dataclasses import dataclass
from dataclasses import field
from mixpanel import BufferedConsumer
from urllib.error import URLError

import json
import structlog
import typing as t

logger = structlog.get_logger(__name__)


@dataclass(frozen=True)
class MockedMessage:
    """A single Mixpanel message stored by MockedConsumer."""

    endpoint: str
    msg: t.Dict["str", object]


@dataclass(frozen=True)
class MockedConsumer:
    """Save messages in an internal list, useful in unit testing."""

    # Internal storage of mocked message
    mocked_messages: t.List = field(default_factory=lambda: [])

    def send(self, endpoint: str, json_message: str) -> None:
        """Append message to the mocked_messages list."""
        msg = MockedMessage(endpoint=endpoint, msg=json.loads(json_message))
        self.mocked_messages.append(msg)


class PoliteBufferedConsumer(BufferedConsumer):
    """Subclass of BufferedConsumer that logs network errors instead of failing.

    Inspired by:
    https://github.com/mixpanel/mixpanel-python/issues/36#issuecomment-72063207
    """

    def flush(self, *args, **kwargs) -> None:
        """Try to send updates to Mixpanel."""
        try:
            super(PoliteBufferedConsumer, self).flush(*args, **kwargs)
        except URLError:
            logger.exception("It seems like Mixpanel is down.", exc_info=True)


@dataclass(frozen=True)
class QueuedConsumer:
    """Queue sending Mixpanel messages in a separate background queue processor."""

    def send(self, endpoint: str, json_message: str) -> None:
        """Queue sending of Mixpanel message in a background task."""
        # send_api.delay(endpoint, json_message)
        raise NotImplementedError  # pragma: no cover



"""Tracking user events and profiles."""

from mixpanel import Mixpanel
from pyramid.events import NewRequest
from pyramid.path import DottedNameResolver
from pyramid.request import Request
from pyramid.response import Response
from pyramid_mixpanel import Event
from pyramid_mixpanel import EventProperties
from pyramid_mixpanel import Events
from pyramid_mixpanel import ProfileMetaProperties
from pyramid_mixpanel import ProfileProperties
from pyramid_mixpanel import Property
from pyramid_mixpanel.consumer import MockedConsumer
from pyramid_mixpanel.consumer import PoliteBufferedConsumer

import typing as t

SettingsType = t.Dict[str, t.Union[str, int, bool]]
PropertiesType = t.Dict[Property, t.Union[str, int, bool]]


def distinct_id_is_required(function: t.Callable) -> t.Callable:
    """Raise AttributeError if self.distinct_id is not set on MixpanelTrack."""

    def wrapper(*args, **kwargs):
        self = args[0]
        if not self.distinct_id:
            raise AttributeError(
                "distinct_id must be set before you can send events or set properties"
            )
        return function(*args, **kwargs)

    return wrapper


class MixpanelTrack:
    """Wrapper around the official `mixpanel` server-side integration for Mixpanel.

    You can track events and/or set people profiles. Uses
    https://pypi.python.org/pypi/mixpanel under the hood.

    Prepared as `request.mixpanel` for easy handling.
    """

    events: Events
    event_properties: EventProperties
    profile_properties: ProfileProperties
    profile_meta_properties: ProfileMetaProperties

    @staticmethod
    def _resolve_events(dotted_name: t.Optional[object] = None) -> Events:
        """Resolve a dotted-name into an Events object and set it to self.events."""
        if not dotted_name:
            return Events()
        if not isinstance(dotted_name, str):
            raise ValueError(
                f"dotted_name must be a string, but it is: {dotted_name.__class__.__name__}"
            )
        else:
            resolved = DottedNameResolver().resolve(dotted_name)
            if not issubclass(resolved, Events):
                raise ValueError(
                    f"class in dotted_name needs to be based on pyramid_mixpanel.Events"
                )
            return resolved()

    @staticmethod
    def _resolve_event_properties(
        dotted_name: t.Optional[object] = None
    ) -> EventProperties:
        """Resolve a dotted-name into an EventProperties object and set it to self.event_properties."""
        if not dotted_name:
            return EventProperties()
        if not isinstance(dotted_name, str):
            raise ValueError(
                f"dotted_name must be a string, but it is: {dotted_name.__class__.__name__}"
            )
        else:
            resolved = DottedNameResolver().resolve(dotted_name)
            if not issubclass(resolved, EventProperties):
                raise ValueError(
                    f"class in dotted_name needs to be based on pyramid_mixpanel.EventProperties"
                )
            return resolved()

    @staticmethod
    def _resolve_profile_properties(
        dotted_name: t.Optional[object] = None
    ) -> ProfileProperties:
        """Resolve a dotted-name into an ProfileProperties object and set it to self.profile_properties."""
        if not dotted_name:
            return ProfileProperties()
        if not isinstance(dotted_name, str):
            raise ValueError(
                f"dotted_name must be a string, but it is: {dotted_name.__class__.__name__}"
            )
        else:
            resolved = DottedNameResolver().resolve(dotted_name)
            if not issubclass(resolved, ProfileProperties):
                raise ValueError(
                    f"class in dotted_name needs to be based on pyramid_mixpanel.ProfileProperties"
                )
            return resolved()

    @staticmethod
    def _resolve_profile_meta_properties(
        dotted_name: t.Optional[object] = None
    ) -> ProfileMetaProperties:
        """Resolve a dotted-name into an ProfileMetaProperties object and set it to self.profile_meta_properties."""
        if not dotted_name:
            return ProfileMetaProperties()
        if not isinstance(dotted_name, str):
            raise ValueError(
                f"dotted_name must be a string, but it is: {dotted_name.__class__.__name__}"
            )
        else:
            resolved = DottedNameResolver().resolve(dotted_name)
            if not issubclass(resolved, ProfileMetaProperties):
                raise ValueError(
                    f"class in dotted_name needs to be based on pyramid_mixpanel.ProfileMetaProperties"
                )
            return resolved()

    def __init__(self, settings: SettingsType, distinct_id=None) -> None:
        """Initialize API connector."""
        self.distinct_id = distinct_id

        if settings.get("mixpanel.token"):
            self.api = Mixpanel(
                token=settings["mixpanel.token"], consumer=PoliteBufferedConsumer()
            )
        else:
            self.api = Mixpanel(token="testing", consumer=MockedConsumer())  # nosec

        self.events = self._resolve_events(settings.get("mixpanel.events"))
        self.event_properties = self._resolve_event_properties(
            settings.get("mixpanel.event_properties")
        )
        self.profile_properties = self._resolve_profile_properties(
            settings.get("mixpanel.profile_properties")
        )
        self.profile_meta_properties = self._resolve_profile_meta_properties(
            settings.get("mixpanel.profile_meta_properties")
        )

    @distinct_id_is_required
    def track(self, event: Event, props: t.Optional[PropertiesType] = None) -> None:
        """Track a Mixpanel event."""
        if not props:
            props = {}

        if event not in self.events.__dict__.values():
            raise ValueError(f"Event '{event}' is not a member of self.events")

        for prop in props.keys():
            if prop not in self.event_properties.__dict__.values():
                raise ValueError(
                    f"Property '{prop}' is not a member of self.event_properties"
                )

        self.api.track(
            self.distinct_id,
            event.name,
            {prop.name: value for (prop, value) in props.items()},
        )

    @distinct_id_is_required
    def profile_set(
        self, props: PropertiesType, meta: t.Optional[PropertiesType] = None
    ) -> None:
        """Set properties to a Profile.

        This creates a profile if one does not yet exist.

        Use `meta` to override are Mixpanel special properties, such as $ip.
        """
        if not meta:
            meta = {}

        for prop in props.keys():
            if prop not in self.profile_properties.__dict__.values():
                raise ValueError(
                    f"Property '{prop}' is not a member of self.profile_properties"
                )

        for prop in meta.keys():
            if prop not in self.profile_meta_properties.__dict__.values():
                raise ValueError(
                    f"Property '{prop}' is not a member of self.profile_meta_properties"
                )

        self.api.people_set(
            self.distinct_id,
            {prop.name: value for (prop, value) in props.items()},
            {prop.name: value for (prop, value) in meta.items()},
        )

    @distinct_id_is_required
    def people_append(
        self, props: PropertiesType, meta: t.Optional[PropertiesType] = None
    ) -> None:
        """Wrap around api.people_append to set distinct_id."""
        if not meta:
            meta = {}

        for prop in props.keys():
            if prop not in self.profile_properties.__dict__.values():
                raise ValueError(
                    f"Property '{prop}' is not a member of self.profile_properties"
                )

        for prop in meta.keys():
            if prop not in self.profile_meta_properties.__dict__.values():
                raise ValueError(
                    f"Property '{prop}' is not a member of self.profile_meta_properties"
                )

        self.api.people_append(
            self.distinct_id,
            {prop.name: value for (prop, value) in props.items()},
            {prop.name: value for (prop, value) in meta.items()},
        )

    @distinct_id_is_required
    def profile_increment(self, props: t.Dict[Property, int]) -> None:
        """Wrap around api.people_increment to set distinct_id."""
        for prop in props.keys():
            if prop not in self.profile_properties.__dict__.values():
                raise ValueError(
                    f"Property '{prop}' is not a member of self.profile_properties"
                )

        self.api.people_increment(
            self.distinct_id, {prop.name: value for (prop, value) in props.items()}
        )

    @distinct_id_is_required
    def profile_track_charge(
        self, amount: int, props: t.Optional[t.Dict[Property, str]] = None
    ) -> None:
        """Wrap around api.people_track_charge to set distinct_id."""
        if not props:
            props = {}

        for prop in props.keys():
            if prop not in self.profile_properties.__dict__.values():
                raise ValueError(
                    f"Property '{prop}' is not a member of self.profile_properties"
                )

        self.api.people_track_charge(
            self.distinct_id,
            amount,
            {prop.name: value for (prop, value) in props.items()},
        )


def mixpanel_init(request: Request) -> MixpanelTrack:
    """Return a configured MixpanelTrack class instance."""
    distinct_id = None
    if getattr(request, "user", None):
        distinct_id = request.user.distinct_id

    return MixpanelTrack(settings=request.registry.settings, distinct_id=distinct_id)


def mixpanel_flush(event: NewRequest) -> None:
    """Send out all pending messages on Pyramid request end."""

    def flush(request: Request, response: Response) -> None:
        """Send all the enqueued messages at the end of request lifecycle."""

        # If request.mixpanel was never called during request runtime, then
        # skip initializing and flushing MixpanelTrack.
        if "mixpanel" not in request.__dict__:
            return

        # Send all messages to Mixpanel for those consumers that require
        # manual flushing at the end of request life-cycle.
        if getattr(request.mixpanel.api._consumer, "flush", None):  # noqa: 236
            request.mixpanel.api._consumer.flush()

    event.request.add_response_callback(flush)



"""Tests for Mixpanel tracking."""

from dataclasses import dataclass
from freezegun import freeze_time
from pyramid_mixpanel import Event
from pyramid_mixpanel import EventProperties
from pyramid_mixpanel import Events
from pyramid_mixpanel import ProfileMetaProperties
from pyramid_mixpanel import ProfileProperties
from pyramid_mixpanel import Property
from pyramid_mixpanel.consumer import MockedConsumer
from pyramid_mixpanel.consumer import PoliteBufferedConsumer
from pyramid_mixpanel.track import MixpanelTrack
from unittest import mock

import pytest


def test_mixpanel_init_distinct_id() -> None:
    """Test distinct_id is set in mixpanel_init function."""
    from pyramid_mixpanel.track import mixpanel_init

    # Requests without request.user
    request = mock.Mock(spec="registry".split())
    request.registry.settings = {}

    result = mixpanel_init(request)

    assert result.__class__ == MixpanelTrack
    assert result.distinct_id is None

    # Requests with request.user
    request = mock.Mock(spec="registry user".split())
    request.registry.settings = {}
    request.user.distinct_id = "foo"

    result = mixpanel_init(request)

    assert result.__class__ == MixpanelTrack
    assert result.distinct_id == "foo"


def test_init_consumers() -> None:
    """Test initialization of Consumer."""
    mixpanel = MixpanelTrack(settings={"mixpanel.token": "secret"})
    assert mixpanel.api._consumer.__class__ == PoliteBufferedConsumer

    mixpanel = MixpanelTrack(settings={})
    assert mixpanel.api._consumer.__class__ == MockedConsumer


@dataclass(frozen=True)
class FooEvents(Events):
    foo: Event = Event("Foo")


@dataclass(frozen=True)
class BarEvents:
    bar: Event = Event("Bar")


def test_init_events() -> None:
    """Test initialization of self.events."""
    # default Events
    mixpanel = MixpanelTrack(settings={})
    assert mixpanel.events == Events()

    # resolved from a dotted-name
    mixpanel = MixpanelTrack(
        settings={"mixpanel.events": "pyramid_mixpanel.tests.test_track.FooEvents"}
    )
    assert mixpanel.events == FooEvents()

    # the resolved Events need to be based off of pyramid_mixpanel.Events
    # to contain the events that this library expects
    with pytest.raises(ValueError) as exc:
        mixpanel = MixpanelTrack(
            settings={"mixpanel.events": "pyramid_mixpanel.tests.test_track.BarEvents"}
        )
    assert (
        str(exc.value)
        == "class in dotted_name needs to be based on pyramid_mixpanel.Events"
    )

    # passing Events as an object is not (yet) supported
    with pytest.raises(ValueError) as exc:
        mixpanel = MixpanelTrack(
            settings={"mixpanel.events": FooEvents()}  # type: ignore
        )
    assert str(exc.value) == "dotted_name must be a string, but it is: FooEvents"


@dataclass(frozen=True)
class FooEventProperties(EventProperties):
    foo: Property = Property("Foo")


@dataclass(frozen=True)
class BarEventProperties:
    bar: Property = Property("Bar")


def test_init_event_properties() -> None:
    """Test initialization of self.event_properties."""
    # default EventProperties
    mixpanel = MixpanelTrack(settings={})
    assert mixpanel.event_properties == EventProperties()

    # resolved from a dotted-name
    mixpanel = MixpanelTrack(
        settings={
            "mixpanel.event_properties": "pyramid_mixpanel.tests.test_track.FooEventProperties"
        }
    )
    assert mixpanel.event_properties == FooEventProperties()

    # the resolved EventProperties need to be based off of
    # pyramid_mixpanel.EventProperties to contain the event properties
    # that this library expects
    with pytest.raises(ValueError) as exc:
        mixpanel = MixpanelTrack(
            settings={
                "mixpanel.event_properties": "pyramid_mixpanel.tests.test_track.BarEventProperties"
            }
        )
    assert (
        str(exc.value)
        == "class in dotted_name needs to be based on pyramid_mixpanel.EventProperties"
    )

    # passing EventProperties as an object is not (yet) supported
    with pytest.raises(ValueError) as exc:
        mixpanel = MixpanelTrack(
            settings={"mixpanel.event_properties": FooEventProperties()}  # type: ignore
        )
    assert (
        str(exc.value) == "dotted_name must be a string, but it is: FooEventProperties"
    )


@dataclass(frozen=True)
class FooProfileProperties(ProfileProperties):
    foo: Property = Property("Foo")


@dataclass(frozen=True)
class BarProfileProperties:
    bar: Property = Property("Bar")


def test_init_profile_properties() -> None:
    """Test initialization of self.profile_properties."""
    # default ProfileProperties
    mixpanel = MixpanelTrack(settings={})
    assert mixpanel.profile_properties == ProfileProperties()

    # resolved from a dotted-name
    mixpanel = MixpanelTrack(
        settings={
            "mixpanel.profile_properties": "pyramid_mixpanel.tests.test_track.FooProfileProperties"
        }
    )
    assert mixpanel.profile_properties == FooProfileProperties()

    # the resolved ProfileProperties need to be based off of
    # pyramid_mixpanel.ProfileProperties to contain the profile properties
    # that this library expects
    with pytest.raises(ValueError) as exc:
        mixpanel = MixpanelTrack(
            settings={
                "mixpanel.profile_properties": "pyramid_mixpanel.tests.test_track.BarProfileProperties"
            }
        )
    assert (
        str(exc.value)
        == "class in dotted_name needs to be based on pyramid_mixpanel.ProfileProperties"
    )

    # passing ProfileProperties as an object is not (yet) supported
    with pytest.raises(ValueError) as exc:
        mixpanel = MixpanelTrack(
            settings={  # type: ignore
                "mixpanel.profile_properties": FooProfileProperties()
            }
        )
    assert (
        str(exc.value)
        == "dotted_name must be a string, but it is: FooProfileProperties"
    )


@dataclass(frozen=True)
class FooProfileMetaProperties(ProfileMetaProperties):
    foo: Property = Property("Foo")


@dataclass(frozen=True)
class BarProfileMetaProperties:
    bar: Property = Property("Bar")


def test_init_profile_meta_properties() -> None:
    """Test initialization of self.profile_meta_properties."""
    # default ProfileMetaProperties
    mixpanel = MixpanelTrack(settings={})
    assert mixpanel.profile_meta_properties == ProfileMetaProperties()

    # resolved from a dotted-name
    mixpanel = MixpanelTrack(
        settings={
            "mixpanel.profile_meta_properties": "pyramid_mixpanel.tests.test_track.FooProfileMetaProperties"
        }
    )
    assert mixpanel.profile_meta_properties == FooProfileMetaProperties()

    # the resolved ProfileMetaProperties need to be based off of
    # pyramid_mixpanel.ProfileMetaProperties to contain the profile properties
    # that this library expects
    with pytest.raises(ValueError) as exc:
        mixpanel = MixpanelTrack(
            settings={
                "mixpanel.profile_meta_properties": "pyramid_mixpanel.tests.test_track.BarProfileMetaProperties"
            }
        )
    assert (
        str(exc.value)
        == "class in dotted_name needs to be based on pyramid_mixpanel.ProfileMetaProperties"
    )

    # passing ProfileMetaProperties as an object is not (yet) supported
    with pytest.raises(ValueError) as exc:
        mixpanel = MixpanelTrack(
            settings={  # type: ignore
                "mixpanel.profile_meta_properties": FooProfileMetaProperties()
            }
        )
    assert (
        str(exc.value)
        == "dotted_name must be a string, but it is: FooProfileMetaProperties"
    )


@freeze_time("2018-01-01")
def test_track() -> None:
    """Test the track method."""
    m = MixpanelTrack(settings={}, distinct_id="foo")

    # default event
    m.track(Events.user_logged_in)
    assert len(m.api._consumer.mocked_messages) == 1
    assert m.api._consumer.mocked_messages[0].endpoint == "events"
    assert m.api._consumer.mocked_messages[0].msg == {
        "event": "User Logged In",
        "properties": {
            "token": "testing",
            "distinct_id": "foo",
            "time": 1514764800,  # 2018-01-01
            "mp_lib": "python",
            "$lib_version": "4.4.0",
        },
    }
    m.api._consumer.mocked_messages.clear()

    # default event with default properties
    m.track(
        Events.page_viewed,
        {
            EventProperties.path: "/about",
            EventProperties.title: "About Us",
            EventProperties.dollar_referrer: "https://niteo.co",
        },
    )
    assert len(m.api._consumer.mocked_messages) == 1
    assert m.api._consumer.mocked_messages[0].endpoint == "events"
    assert m.api._consumer.mocked_messages[0].msg == {
        "event": "Page Viewed",
        "properties": {
            "token": "testing",
            "distinct_id": "foo",
            "time": 1514764800,  # 2018-01-01
            "mp_lib": "python",
            "$lib_version": "4.4.0",
            "$referrer": "https://niteo.co",
            "Path": "/about",
            "Title": "About Us",
        },
    }
    m.api._consumer.mocked_messages.clear()

    # custom event with custom properties
    m = MixpanelTrack(
        settings={
            "mixpanel.events": "pyramid_mixpanel.tests.test_track.FooEvents",
            "mixpanel.event_properties": "pyramid_mixpanel.tests.test_track.FooEventProperties",
        },
        distinct_id="foo",
    )
    m.track(FooEvents.foo, {FooEventProperties.foo: "bar"})
    assert len(m.api._consumer.mocked_messages) == 1
    assert m.api._consumer.mocked_messages[0].endpoint == "events"
    assert m.api._consumer.mocked_messages[0].msg == {
        "event": "Foo",
        "properties": {
            "token": "testing",
            "distinct_id": "foo",
            "time": 1514764800,
            "mp_lib": "python",
            "$lib_version": "4.4.0",
            "Foo": "bar",
        },
    }

    m.api._consumer.mocked_messages.clear()


def test_track_guards() -> None:
    """Test guards that make sure parameters sent to .track() are good."""

    # fail if distinct_id is None
    m = MixpanelTrack(settings={})
    with pytest.raises(AttributeError) as exc:
        m.track(Events.user_logged_in)
    assert (
        str(exc.value)
        == "distinct_id must be set before you can send events or set properties"
    )

    # fail if event is not a member of mixpanel.events
    m = MixpanelTrack(settings={}, distinct_id="foo")
    with pytest.raises(ValueError) as exc:
        m.track(FooEvents.foo)
    assert str(exc.value) == "Event 'Event(name='Foo')' is not a member of self.events"

    # fail if event property is not a member of mixpanel.event_properties
    m = MixpanelTrack(settings={}, distinct_id="foo")
    with pytest.raises(ValueError) as exc:
        m.track(Events.user_logged_in, {FooEventProperties.foo: "foo"})
    assert (
        str(exc.value)
        == "Property 'Property(name='Foo')' is not a member of self.event_properties"
    )


@freeze_time("2018-01-01")
def test_profile_set() -> None:
    """Test the profile_set method."""
    m = MixpanelTrack(settings={}, distinct_id="foo")

    m.profile_set({ProfileProperties.dollar_name: "FooBar"})
    assert len(m.api._consumer.mocked_messages) == 1
    assert m.api._consumer.mocked_messages[0].endpoint == "people"
    assert m.api._consumer.mocked_messages[0].msg == {
        "$token": "testing",
        "$time": 1514764800000,
        "$distinct_id": "foo",
        "$set": {"$name": "FooBar"},
    }

    # with meta properties
    m.profile_set(
        {ProfileProperties.dollar_name: "FooBar2"},
        meta={ProfileMetaProperties.dollar_ip: "1.1.1.1"},
    )
    assert len(m.api._consumer.mocked_messages) == 2
    assert m.api._consumer.mocked_messages[1].endpoint == "people"
    assert m.api._consumer.mocked_messages[1].msg == {
        "$token": "testing",
        "$time": 1514764800000,
        "$distinct_id": "foo",
        "$set": {"$name": "FooBar2"},
        "$ip": "1.1.1.1",
    }


def test_profile_set_guards() -> None:
    """Test guards that make sure parameters sent to .profile_set() are good."""

    # fail if distinct_id is None
    m = MixpanelTrack(settings={})
    with pytest.raises(AttributeError) as exc:
        m.profile_set({ProfileProperties.dollar_name: "FooBar"})
    assert (
        str(exc.value)
        == "distinct_id must be set before you can send events or set properties"
    )

    # fail if property is not a member of mixpanel.profile_properties
    m = MixpanelTrack(settings={}, distinct_id="foo")
    with pytest.raises(ValueError) as exc:
        m.profile_set({FooProfileProperties.foo: "bar"})
    assert (
        str(exc.value)
        == "Property 'Property(name='Foo')' is not a member of self.profile_properties"
    )

    # fail if meta property is not a member of mixpanel.profile_meta_properties
    m = MixpanelTrack(settings={}, distinct_id="foo")
    with pytest.raises(ValueError) as exc:
        m.profile_set(
            {ProfileProperties.dollar_name: "foo"},
            meta={FooProfileMetaProperties.foo: "bar"},
        )
    assert (
        str(exc.value)
        == "Property 'Property(name='Foo')' is not a member of self.profile_meta_properties"
    )


@freeze_time("2018-01-01")
def test_people_append() -> None:
    """Test the people_append method."""
    m = MixpanelTrack(settings={}, distinct_id="foo")

    m.people_append({ProfileProperties.dollar_name: "FooBar"})
    assert len(m.api._consumer.mocked_messages) == 1
    assert m.api._consumer.mocked_messages[0].endpoint == "people"
    assert m.api._consumer.mocked_messages[0].msg == {
        "$token": "testing",
        "$time": 1514764800000,
        "$distinct_id": "foo",
        "$append": {"$name": "FooBar"},
    }

    # with meta properties
    m.people_append(
        {ProfileProperties.dollar_name: "FooBar2"},
        meta={ProfileMetaProperties.dollar_ip: "1.1.1.1"},
    )
    assert len(m.api._consumer.mocked_messages) == 2
    assert m.api._consumer.mocked_messages[1].endpoint == "people"
    assert m.api._consumer.mocked_messages[1].msg == {
        "$token": "testing",
        "$time": 1514764800000,
        "$distinct_id": "foo",
        "$append": {"$name": "FooBar2"},
        "$ip": "1.1.1.1",
    }


def test_people_append_guards() -> None:
    """Test guards that make sure parameters sent to .people_append() are good."""

    # fail if distinct_id is None
    m = MixpanelTrack(settings={})
    with pytest.raises(AttributeError) as exc:
        m.people_append({ProfileProperties.dollar_name: "FooBar"})
    assert (
        str(exc.value)
        == "distinct_id must be set before you can send events or set properties"
    )

    # fail if property is not a member of mixpanel.profile_properties
    m = MixpanelTrack(settings={}, distinct_id="foo")
    with pytest.raises(ValueError) as exc:
        m.people_append({FooProfileProperties.foo: "FooBar"})
    assert (
        str(exc.value)
        == "Property 'Property(name='Foo')' is not a member of self.profile_properties"
    )

    # fail if meta property is not a member of mixpanel.profile_meta_properties
    m = MixpanelTrack(settings={}, distinct_id="foo")
    with pytest.raises(ValueError) as exc:
        m.people_append(
            {ProfileProperties.dollar_name: "foo"},
            meta={FooProfileMetaProperties.foo: "bar"},
        )
    assert (
        str(exc.value)
        == "Property 'Property(name='Foo')' is not a member of self.profile_meta_properties"
    )


@freeze_time("2018-01-01")
def test_profile_increment() -> None:
    """Test the profile_increment method."""
    m = MixpanelTrack(
        settings={
            "mixpanel.profile_properties": "pyramid_mixpanel.tests.test_track.FooProfileProperties"
        },
        distinct_id="foo",
    )

    m.profile_increment(props={FooProfileProperties.foo: 1})
    assert len(m.api._consumer.mocked_messages) == 1
    assert m.api._consumer.mocked_messages[0].endpoint == "people"
    assert m.api._consumer.mocked_messages[0].msg == {
        "$token": "testing",
        "$time": 1514764800000,
        "$distinct_id": "foo",
        "$add": {"Foo": 1},
    }


def test_profile_increment_guards() -> None:
    """Test guards that make sure parameters sent to .profile_increment() are good."""

    # fail if distinct_id is None
    m = MixpanelTrack(settings={})
    with pytest.raises(AttributeError) as exc:
        m.profile_increment({ProfileProperties.dollar_name: "FooBar"})
    assert (
        str(exc.value)
        == "distinct_id must be set before you can send events or set properties"
    )

    # fail if property is not a member of mixpanel.profile_properties
    m = MixpanelTrack(settings={}, distinct_id="foo")
    with pytest.raises(ValueError) as exc:
        m.profile_increment({FooProfileProperties.foo: "FooBar"})
    assert (
        str(exc.value)
        == "Property 'Property(name='Foo')' is not a member of self.profile_properties"
    )


@freeze_time("2018-01-01")
def test_profile_track_charge() -> None:
    """Test the profile_track_charge method."""
    m = MixpanelTrack(
        settings={
            "mixpanel.profile_properties": "pyramid_mixpanel.tests.test_track.FooProfileProperties"
        },
        distinct_id="foo",
    )

    m.profile_track_charge(100)
    assert len(m.api._consumer.mocked_messages) == 1
    assert m.api._consumer.mocked_messages[0].endpoint == "people"
    assert m.api._consumer.mocked_messages[0].msg == {
        "$token": "testing",
        "$time": 1514764800000,
        "$distinct_id": "foo",
        "$append": {"$transactions": {"$amount": 100}},
    }

    m.profile_track_charge(222, props={FooProfileProperties.foo: "Bar"})
    assert len(m.api._consumer.mocked_messages) == 2
    assert m.api._consumer.mocked_messages[1].endpoint == "people"
    assert m.api._consumer.mocked_messages[1].msg == {
        "$token": "testing",
        "$time": 1514764800000,
        "$distinct_id": "foo",
        "$append": {"$transactions": {"Foo": "Bar", "$amount": 222}},
    }


def test_profile_track_charge_guards() -> None:
    """Test guards that make sure parameters sent to .profile_track_charge() are good."""

    # fail if distinct_id is None
    m = MixpanelTrack(settings={})
    with pytest.raises(AttributeError) as exc:
        m.profile_track_charge(100, {ProfileProperties.dollar_name: "FooBar"})
    assert (
        str(exc.value)
        == "distinct_id must be set before you can send events or set properties"
    )

    # fail if property is not a member of mixpanel.profile_properties
    m = MixpanelTrack(settings={}, distinct_id="foo")
    with pytest.raises(ValueError) as exc:
        m.profile_track_charge(100, {FooProfileProperties.foo: "FooBar"})
    assert (
        str(exc.value)
        == "Property 'Property(name='Foo')' is not a member of self.profile_properties"
    )



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from io import open
from setuptools import find_packages
from setuptools import setup
from setuptools.command.install import install

import os
import sys

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(os.path.join(here, "CHANGELOG.md"), encoding="utf-8") as f:
    long_description += "\n\n" + f.read()


VERSION = "0.4.3"


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version.

    Taken from https://circleci.com/blog/continuously-deploying-python-packages-to-pypi-with-circleci/
    """

    description = "verify that the git tag matches our version"

    def run(self):  # noqa: D102
        tag = os.getenv("CIRCLE_TAG")

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


setup(
    name="pyramid_mixpanel",
    version=VERSION,
    description="Opinionated Pyramid integration with Mixpanel, a user behavioural analytics platform and CRM.",
    long_description=long_description,
    license="MIT",
    long_description_content_type="text/markdown",
    url="https://github.com/niteoweb/pyramid_mixpanel",
    author="Niteo",
    author_email="info@niteo.co",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Pyramid",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="pyramid mixpanel pylons web",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    install_requires=["pyramid", "requests", "structlog", "mixpanel"],
    cmdclass={"verify": VerifyVersionCommand},
)



"""Functional tests against a real Pyramid app."""

from freezegun import freeze_time
from mixpanel import json_dumps
from pyramid.config import Configurator
from pyramid.request import Request
from pyramid.router import Router
from pyramid.view import view_config
from pyramid_mixpanel import EventProperties
from pyramid_mixpanel import Events
from pyramid_mixpanel.consumer import MockedMessage
from testfixtures import LogCapture
from unittest import mock
from webtest import TestApp

import base64
import structlog
import typing as t
import urllib


@view_config(route_name="hello", renderer="json", request_method="GET")
def hello(request: Request) -> t.Dict[str, str]:
    """Say hello."""
    # mocking that request has a user object
    request.user = mock.MagicMock(spec="distinct_id".split())
    request.user.distinct_id = "foo-123"

    # provide access to Pyramid request in WebTest response
    request.environ["paste.testing_variables"]["app_request"] = request

    request.mixpanel.track(Events.page_viewed, {EventProperties.path: "/hello"})
    return {"hello": "world"}


@view_config(route_name="bye", renderer="json", request_method="GET")
def bye(request: Request) -> t.Dict[str, str]:
    """Say bye."""
    return {"bye": "bye"}


def app(settings) -> Router:
    """Create a dummy Pyramid app."""
    structlog.configure(
        processors=[structlog.processors.KeyValueRenderer(sort_keys=True)],
        logger_factory=structlog.stdlib.LoggerFactory(),
    )

    with Configurator() as config:
        config.add_route("hello", "/hello")
        config.add_route("bye", "/bye")
        config.scan(".")

        config.registry.settings.update(**settings)
        config.include("pyramid_mixpanel")

        return config.make_wsgi_app()


@freeze_time("2019-01-01")
def test_MockedConsumer() -> None:
    """Test that request.mixpanel works as expected with MockedConsumer."""
    with LogCapture() as logs:
        testapp = TestApp(app({}))

        # do two requests to make sure logs are not flooded with messages
        # on every request
        res = testapp.get("/hello", status=200)
        res = testapp.get("/hello", status=200)
        assert res.json == {"hello": "world"}

    logs.check(
        (
            "pyramid_mixpanel",
            "INFO",
            "consumer='MockedConsumer' event='Mixpanel configured' "
            "event_properties='EventProperties' events='Events' "
            "profile_meta_properties='ProfileMetaProperties' "
            "profile_properties='ProfileProperties'",
        ),
        (
            "pyramid_mixpanel",
            "WARNING",
            "event='Mixpanel is in testing mode, no message will be sent!'",
        ),
    )

    res.app_request.mixpanel.api._consumer.mocked_messages == [
        MockedMessage(
            endpoint="events",
            msg={
                "event": "Page Viewed",
                "properties": {
                    "token": "testing",
                    "distinct_id": "foo-123",
                    "time": 1564322909,
                    "mp_lib": "python",
                    "$lib_version": "4.4.0",
                    "Path": "/hello",
                },
            },
        )
    ]


@freeze_time("2019-01-01")
@mock.patch("mixpanel.urllib.request.urlopen")
@mock.patch("mixpanel.urllib.request.Request")
def test_PoliteBufferedConsumer(
    request: mock.MagicMock, urlopen: mock.MagicMock
) -> None:
    """Test that request.mixpanel works as expected with PoliteBufferedConsumer."""
    urlopen().read.return_value = b'{"error":null,"status":1}'

    with LogCapture() as logs:
        settings = {"mixpanel.token": "SECRET"}
        testapp = TestApp(app(settings))

        res = testapp.get("/hello", status=200)
        assert res.json == {"hello": "world"}

    logs.check(
        (
            "pyramid_mixpanel",
            "INFO",
            "consumer='PoliteBufferedConsumer' event='Mixpanel configured' "
            "event_properties='EventProperties' events='Events' "
            "profile_meta_properties='ProfileMetaProperties' "
            "profile_properties='ProfileProperties'",
        )
    )

    message = json_dumps(
        [
            {
                "event": "Page Viewed",
                "properties": {
                    "token": "SECRET",
                    "distinct_id": "foo-123",
                    "time": 1546300800,
                    "mp_lib": "python",
                    "$lib_version": "4.4.0",
                    "Path": "/hello",
                },
            }
        ]
    )
    data = {"data": base64.b64encode(message.encode("utf8")), "verbose": 1, "ip": 0}
    request.assert_called_with(
        "https://api.mixpanel.com/track",
        urllib.parse.urlencode(data).encode("utf8"),  # type:ignore
    )


@mock.patch("pyramid_mixpanel.consumer.PoliteBufferedConsumer.flush")
def test_request_mixpanel_not_used(flush: mock.MagicMock) -> None:
    """Test that flush() is not called if request.mixpanel was never called."""

    with LogCapture() as logs:
        settings = {"mixpanel.token": "SECRET"}
        testapp = TestApp(app(settings))

        res = testapp.get("/bye", status=200)
        assert res.json == {"bye": "bye"}

    logs.check(
        (
            "pyramid_mixpanel",
            "INFO",
            "consumer='PoliteBufferedConsumer' event='Mixpanel configured' "
            "event_properties='EventProperties' events='Events' "
            "profile_meta_properties='ProfileMetaProperties' "
            "profile_properties='ProfileProperties'",
        )
    )
    flush.assert_not_called()



"""Tests for provided consumers."""

from pyramid_mixpanel.consumer import MockedConsumer
from pyramid_mixpanel.consumer import MockedMessage
from pyramid_mixpanel.consumer import PoliteBufferedConsumer
from testfixtures import LogCapture
from unittest import mock
from urllib.error import URLError

import structlog


def test_MockedConsumer() -> None:
    """Test that MockedConsumer saves messages."""
    consumer = MockedConsumer()

    consumer.send(endpoint="events", json_message='{"foo":"Foo"}')
    consumer.send(endpoint="events", json_message='{"bar":"Bar"}')

    assert len(consumer.mocked_messages) == 2
    assert consumer.mocked_messages[0] == MockedMessage("events", {"foo": "Foo"})
    assert consumer.mocked_messages[1] == MockedMessage("events", {"bar": "Bar"})


@mock.patch("mixpanel.BufferedConsumer.flush")
def test_PoliteBufferedConsumer(flush: mock.MagicMock) -> None:
    """Test that PoliteBufferedConsumer logs errors and continues."""
    structlog.configure(
        processors=[structlog.processors.KeyValueRenderer(sort_keys=True)],
        logger_factory=structlog.stdlib.LoggerFactory(),
    )

    consumer = PoliteBufferedConsumer()

    consumer.send(endpoint="events", json_message='{"foo":"Foo"}')
    consumer.send(endpoint="events", json_message='{"bar":"Bar"}')

    assert consumer._buffers == {  # noqa: SF01
        "events": ['{"foo":"Foo"}', '{"bar":"Bar"}'],
        "people": [],
        "imports": [],
    }

    consumer.flush()
    flush.assert_called_with()

    with LogCapture() as logs:
        flush.side_effect = URLError("foo")
        consumer.flush()

    logs.check(
        (
            "pyramid_mixpanel.consumer",
            "ERROR",
            "event='It seems like Mixpanel is down.' exc_info=True",
        )
    )
